#!/usr/bin/perl
## ######### PROJECT NAME : ##########
## 
## LibDisk.pm
## 
## ######### PROJECT DESCRIPTION : ###
## 
## This library permits to developpers
## to handle disks and partitions easily.
##
## This library is based on lib_infodisks.pm
## originally designed by Sardinha Mickael.
## 
## ###################################
## 
## Made by Boutonnet Alexandre
## Login   <aboutonnet@intellique.com>
## 
## Started on  Mon Feb 26 15:54:54 2007 Sardinha Mickael
## Last update Tue Jun  9 11:31:50 2009 Boutonnet Alexandre
## 
## ###################################
##

package INTLIQ::LibDisk;

# Declaration des Modules utilises
use strict;
use warnings;
use Data::Dumper;

# Definition des binaires utilises
use constant DEV => "/dev/";

my $known_fs = {
		"primary" => undef,
		"logical" => undef,
		"extended" => undef,
		"primaire" => undef,
		"logique" => undef,
		"tendue" => undef,
		"" => undef,
		};

my $known_label = {
    'gpt' => undef,
    'msdos' => undef, # normalement desactive car on gere que le gpt
};

my $known_filesystem = {
    "reiserfs" => "/sbin/mkfs.reiserfs -q",
    "xfs" => "/sbin/mkfs.xfs -f -q",
    "ext2" => "/sbin/mkfs.ext2 -q",
    "ext3" => "/sbin/mkfs.ext3 -q",
    "vfat" => "/sbin/mkfs.vfat",
    "linux-swap" => "/sbin/mkswap -f",
};

# On set la variable avec le path de fdisk
my $cmd_fdisk		= _test_cmd("fdisk", "/sbin/fdisk");
my $cmd_parted		= _test_cmd("parted","/sbin/parted");

my $error_parted	= "Error: Wrong version of Parted. \n";

my $regexp_md	= qr!(?:\/dev\/md\d|\/dev\/dm-\d|\/dev\/loop\d)!;

my $regex	= qr!^\s*(\d+)\s+(\d+)[Bk]\s+(\d+)B\s+(\d+)B\s+\w+\s+([\-0-9a-zA-Z_]*)\s*! ;
my $regex2	= qr!^\s*(\d+)\s+(\d+)B\s+(\d+)B\s+(\d+)B\s+([\-0-9a-zA-Z_]*)\s*! ;
my $regex3	= qr!^\s*(\d+)\s+(\d+)B\s+(\d+)B\s+(\d+)B\s+([\-0-9a-zA-Z_]*)\s*! ;

# Acquisition de certaines donnees grace a fdisk qu'on place ensuite dans la
# Hash: type donnees - disk(ex hda) - heads - sectors - cylinders
sub get_disk_info
{
    my @fdisk_all = qx!$cmd_fdisk -l 2>/dev/null | grep Di!;
    my $hash = {};
    
    chomp ( my $version_parted = qx!$cmd_parted -v | cut -d " " -f 3 | cut -d "." -f 1-2! );

    # Recuperation de la liste des disques.
    my ($ret_err, $ret_disk) = get_disk_list();
    return ($ret_err, $ret_disk) if ($ret_err);

    foreach my $disk ( @{$ret_disk} )
    {
	_get_fdisk_info($hash, $disk, $version_parted);
	my $ret_part;
	($ret_err, $ret_part) = get_disk_partition($disk);

	if (!$ret_err)
	{
	    my @info_parted = qx!$cmd_parted -s $disk unit B print 2>&1!;
	    _get_parted_info($hash, \@info_parted, $disk, $version_parted);
	}
	else
	{
	    $hash->{$disk}->{'partition'} = 'none';
	}
    }

    return (0, $hash);
}

# Fonction de recuperation de la liste des disques.
# __Attention__ cette methode passe outre les devices /dev/md*
sub get_disk_list
{
    my @fdisk_all	= qx!$cmd_fdisk -l 2>/dev/null | grep Di!;

    my @return_tab;
    foreach my $line (@fdisk_all)
    {
        #Disk /dev/sda: 249.9 GB, 249988907008 bytes
	if ( $line =~ /^\s*(?:Disk|Disque)\s+\/dev\/(.*):\s+.*\s+(\d*)\s+/)
	{
	    push @return_tab, DEV.$1;
	}
    }

    return (0, \@return_tab);
}

# Fonction de recuperation des partitions en fonction d'un disque
# Le parametre $dev est le device sur lequel parted travaille
# S'il est omis, cette methode retourne une erreur.
# Si l'on passe une partition a cette fontion, elle considere la partition
# comme un disque : cela entraine un bug silencieux
sub get_disk_partition
{
    my ($dev) = @_;
    
    return (1, "Device parameter is missing") if (!$dev);
    
    my @return_tab;
    my @info_parted = qx!$cmd_parted -s $dev unit B print 2>&1!;

    # Ici je test si le retour de parted n'est pas une erreur
    # si c'est le cas je quitte en retournant zero
    return(1, "Parted error") if ( !@info_parted || _test_parted_error(@info_parted));
    
    my $flag = 0;
    my $regexp = 0;
    
    foreach my $line (@info_parted) {
	unless ($flag) {
	    if ($line =~ /^\s*(?:Number|Num.ro)/)
	    {
		($flag, $regexp) = _detect_parted_version($line);
		return (1, "Parted version not supported") if (!$flag);
	    }
	}
	if ( $flag && ($line =~ /$regexp/) )
	{
	    if ($dev =~ m/$regexp_md/)
	    {
		push(@return_tab, $dev);
	    }
	    else
	    {
		push(@return_tab, $dev.$1);
	    }
	}
    }
    
    return (0, \@return_tab);
}

# Fonction de recuperation des infos propres a une partition
# Cette methode retourne une reference sur une hash
# Le parametre $part est le nom de la partition. S'il est absent, la fonction
# retourne une erreur
sub get_partition_info
{
    my ($part) = @_;

    return (1, "Partition parameter is missing") if (!$part);

    my ($dev) = ($part =~ m/^(.*)\d$/);
    my ($num_part) = ($part =~ m/^.*(\d)$/);

    if ($part =~ m/$regexp_md/)
    {
	$dev = $part;
	$num_part = 1;
    }

    my $return_hash = {};
    my @info_parted = qx!$cmd_parted -s $dev unit B print 2>&1!;

    # Ici je test si le retour de parted n'est pas une erreur
    # si c'est le cas je quitte en retournant zero
    return(1, "Parted error") if ( !@info_parted || _test_parted_error(@info_parted));

    my $flag = 0;
    my $regexp = 0;
    
    foreach my $line (@info_parted)
    {
	unless ($flag)
	{
	    if ($line =~ /^\s*(?:Number|Num.ro)/)
	    {
		($flag, $regexp) = _detect_parted_version($line);
		return (0) if (!$flag);
	    }
	}
	if ( $flag && ($line =~ /$regexp/) && $1 == $num_part)
	{
	    $return_hash->{$part}{'start'}	= $2 ;
	    $return_hash->{$part}{'end'}	= $3 ;
	    $return_hash->{$part}{'size'}	= $4 ;
	    
	    if ( exists $known_fs->{$5} )
	    {
		$return_hash->{$part}{'filesystem'}	= "Unspecified filesystem" ;
	    }
	    else
	    {
		$return_hash->{$part}{'filesystem'}	= $5 ;
	    }
	}
    }

    return (1, "Unable to find partition size") if (!exists($return_hash->{$part}->{'size'}));

    return (0, $return_hash);
}

# Fonction de recuperation des infos detaillees sur un disque
# Cette methode prend en parametre un device et retourne une erreur dans deux cas :
# - le device n'existe pas
# - le parametre est absent
# Sinon elle retourne une reference sur une hash
sub get_one_disk_info
{
    my ($disk) = @_;
    
    return (1, "Disk parameter is missing") if (!$disk);

    my @fdisk_all = qx!$cmd_fdisk -l $disk 2>/dev/null | grep Di!;

    return (1, "Unable to get disk infos") if (!@fdisk_all);

    my $hash = {};
    
    chomp ( my $version_parted = qx!$cmd_parted -v | cut -d " " -f 3 | cut -d "." -f 1-2! );

    _get_fdisk_info($hash, $disk, $version_parted);
    my ($ret_err, $ret_part) = get_disk_partition($disk);
    
    if (!$ret_err)
    {
	my @info_parted = qx!$cmd_parted -s $disk unit B print 2>&1!;
	_get_parted_info($hash, \@info_parted, $disk, $version_parted);
    }
    else
    {
	$hash->{$disk}->{'partition'} = 'none';
    }
    
    return (0, $hash);
}

# Fonction de partitionnement d'un disque dur..
# Cette fonction prend une hash en parametre.
# Le formattage de la hash est le meme que la sortie de la fonction
# get_disk_info
sub part_disk
{
    my ($hash) = @_;
    
    foreach my $disk (sort keys(%{$hash}))
    {
	# Si le label existe, 
	if (exists($hash->{$disk}->{'label'}))
	{
	    # J'appelle la fonction qui applique le label
	    my ($error, $msg) = mk_disk_label($disk, $hash->{$disk}->{'label'});
	    return ($error, $msg) if ($error);
	}

	# Je teste s'il y a des partitions a faire
	if (exists($hash->{$disk}->{'partition'}))
	{
	    foreach my $part (sort keys(%{$hash->{$disk}->{'partition'}}))
	    {
		# Je cre un hash temporaire..
		my $tmp_hash = {
		    $part => $hash->{$disk}->{'partition'}->{$part},
		};
		# Si start existe, j'appelle la methode de creation des partitions
		if (exists($tmp_hash->{$part}->{'start'}))
		{
		    my ($error, $msg) = mk_disk_partition($disk, $tmp_hash);
		    return ($error, $msg) if ($error);
		}
		# Si j'ai un file system j'appelle la methode de formattage
		if (exists($tmp_hash->{$part}->{'filesystem'}))
		{
		    # Je dors 1 secondes pour viter un problem de stat()
		    # si la partition vient d'tre cr
		    sleep 1;
		    my ($error, $msg) = format_partition($part, $tmp_hash->{$part}->{'filesystem'});
		    return ($error, $msg) if ($error);
		}
	    }
	}
    }
    return (0, '');
}

# Fonction de creation de partition...
# Cette fonction prend en parametre le nom du disque
# ainsi qu'une hash de ce type :
# {
#   '/dev/hdd1' => {
#                   'filesystem' => 'reiserfs',
#                   'start' => '17408',
#                   'size' => '4294982656',
#                   'end' => '4295000063',
#                  },
# }
sub mk_disk_partition
{
    my ($device, $hash) = @_;

    foreach my $part (sort keys(%{$hash}))
    {
	# Je test si j'ai tout ce qui me faut pour faire une partition..
	return (1, "Start value is not defined") if (!exists($hash->{$part}->{'start'}));
	return (1, "End value is not defined") if (!exists($hash->{$part}->{'end'}));

	my @info_parted = qx!$cmd_parted -s $device mkpart -- primary $hash->{$part}->{'start'} $hash->{$part}->{'end'} 2>&1!;
	# my @info_parted = qx!$cmd_parted -s $device unit B mkpart -- primary $hash->{$part}->{'start'} $hash->{$part}->{'end'} 2>&1!;

	# Ici je test si le retour de parted n'est pas une erreur
	# si c'est le cas je quitte en retournant ligne contenant l'erreur
	if (@info_parted && (my $pos = _test_parted_error(@info_parted)))
	{
	    chomp($info_parted[$pos - 1]);
	    return (1, ($info_parted[$pos - 1]));
	}

	# Ici je test si le retour de parted n'est pas un avertissement
	# si c'est le cas je quitte en retournant le message au complet
	if (@info_parted && (my $pos = _test_parted_warning(@info_parted)))
	{
	    my $buf;
	    foreach my $line (@info_parted)
	    {
		$buf .= $line;
	    }
	    return (1, $buf);
	}
    }

    return (0, '');
}

# Fonction de formatage d'un partition
# Cette fonction prend en parametre le device
# ainsi que le file system a appliquer
sub format_partition
{
    my ($device, $filesystem) = @_;

    # Avant de partitionner je teste si j'ai pas un filesystem en bois
    return (1, "Unknown Filesystem for ".$device) if (!exists($known_filesystem->{$filesystem}));

    my $retour_mkfs = system("$known_filesystem->{$filesystem} $device > /dev/null 2>/dev/null");
    return (1, "An error occured during the format process of ".$device) if ($retour_mkfs);

    return (0, '');
}

# Fonction d'application d'un label sur un disque.
# Cette fonction prend en parametre le nom du device
# et le label
sub mk_disk_label
{
    my ($device, $label) = @_;

    return (1, "Disk label unknown") if (!exists($known_label->{$label}));

    my @info_parted = qx!$cmd_parted -s $device mklabel $label 2>&1!;

    # Ici je test si le retour de parted n'est pas une erreur
    # si c'est le cas je quitte en retournant ligne contenant l'erreur
    if (@info_parted && (my $pos = _test_parted_error(@info_parted)))
    {
	chomp($info_parted[$pos - 1]);
	return (1, ($info_parted[$pos - 1]));
    }

    return (0, '');
}

# Fonction de suppression d'une partition
# Cette fonction prend en parametre le nom de la partition
# a supprimer.
sub delete_partition
{
    my ($part) = @_;

    # Je recupere le minor de la partition
    my ($minor) = ($part =~ m/.*(\d+)$/);

    return (1, "Unable to get partition minor for ".$part) if (!$minor);

    my ($device) = ($part =~ m/(.*)\d+$/);

    my @info_parted = qx!$cmd_parted -s $device rm $minor 2>&1!;

    # Ici je test si le retour de parted n'est pas une erreur
    # si c'est le cas je quitte en retournant ligne contenant l'erreur
    if (@info_parted && (my $pos = _test_parted_error(@info_parted)))
    {
	chomp($info_parted[$pos - 1]);
	return (1, ($info_parted[$pos - 1]));
    }
    
    return (0, '');
}

######################## Methodes privees ######################################

# Acquisition de certaines donnees grace a parted qu'on place ensuite dans la
# Hash: type donnees - partition (ex hda1) - start - end - filesystem
sub	_get_parted_info
{
    my $regexp	= 0;
    my $flag	= 0;

    my ($hash, $parted, $disk, $version_parted) = @_;

    foreach my $line (@{$parted})
    {
	unless ($flag)
	{
	    # Detection de la version puis selection de la regexp correspondante
	    if ($line =~ /^\s*(?:Number|Num.ro)/)
	    {
		($flag, $regexp) = _detect_parted_version($line);
		return (0) if (!$flag);
	    }
	}
	#                 2       3800.000   4800.000  linux-swap

	if ( $flag && ($line =~ /$regexp/))
	{
	    if ($disk !~ m/$regexp_md/)
	    {
		$hash->{$disk}{'partition'}{$disk.$1}{'start'} = $2 ;
		$hash->{$disk}{'partition'}{$disk.$1}{'end'} = $3 ;
		$hash->{$disk}{'partition'}{$disk.$1}{'size'} = $4 ;
		
		if ( exists $known_fs->{$5} )
		{
		    $hash->{$disk}{'partition'}{$disk.$1}{'filesystem'} = "Unspecified filesystem" ;
		}
		else
		{
		    $hash->{$disk}{'partition'}{$disk.$1}{'filesystem'} = $5 ;
		}
	    }
	    else
	    {
		# Make captures again..
		($line =~ /$regexp/);

		$hash->{$disk}{'partition'}{$disk}{'start'} = $2 ;
		$hash->{$disk}{'partition'}{$disk}{'end'} = $3 ;
		$hash->{$disk}{'partition'}{$disk}{'size'} = $4 ;
		
		if ( exists $known_fs->{$5} )
		{
		    $hash->{$disk}{'partition'}{$disk}{'filesystem'} = "Unspecified filesystem" ;
		}
		else
		{
		    $hash->{$disk}{'partition'}{$disk}{'filesystem'} = $5 ;
		}
	    }
	}
    }
}

# Acquisition de certaines de donnees grace a parted qu'on place ensuite dans la
# Hash: type donnees - partition (ex hda1) - start - end - filesystem
sub	_get_parted_info_md
{
    my $regexp	= 0;
    my $flag	= 0;

    my ($hash, $parted, $disk, $version_parted) = @_;

    foreach my $line (@{$parted}) {
	unless ($flag) {
	    # Detection de la version puis selection de la regexp correspondante
	    if ($line =~ /^\s*(?:Number|Num.ro)/)	{
		($flag, $regexp) = _detect_parted_version($line);
		return (0) if (!$flag);
	    }
	}
	#                 2       3800.000   4800.000  linux-swap
	if ( $flag && ($line =~ /$regexp/) ) {
	    $hash->{'raid '.$disk}{'partition'}{$disk}{'start'}	= $2 ;
	    $hash->{'raid '.$disk}{'partition'}{$disk}{'end'}	= $3 ;
	    $hash->{'raid '.$disk}{'partition'}{$disk}{'size'}	= $4 ;
#			$hash->{'raid '.$disk}{'size'}						= $4;
	    
	    if ( exists $known_fs->{$5} ) {
		$hash->{'raid '.$disk}{'partition'}{$disk}{'filesystem'}	= "Unspecified filesystem" ;
	    }
	    else {
		$hash->{'raid '.$disk}{'partition'}{$disk}{'filesystem'}	= $5 ;
	    }
	}
    }
}

# Fonction de rcupration des informations de fdisk.
sub _get_fdisk_info
{
    my $flag1 = 0;
    my $flag2 = 0;
    my ($hash, $disk, $version_parted) = @_;
    my @fdisk = qx!$cmd_fdisk -l $disk 2>/dev/null!;
    
    foreach my $line (@fdisk)
    {
	#Disk /dev/sda: 249.9 GB, 249988907008 bytes
	if ( $line =~ /^\s*(?:Disk|Disque)\s+\/dev\/(.*):\s+.*\s+(\d*)\s+/)
	{
	    # Si on n'obtient pas la taille par fdisk, on la recupere dans /proc/partitions
	    if ( $2 == 0)
	    {
		my $temp = qx!grep $1\$ /proc/partitions !;
		my (undef,undef,$real_size) = split (' ', $temp); 
		$hash->{$disk}{'size'} = $real_size;
		$flag1 = 1 ;
	    }
	    else
	    {
		$hash->{$disk}{'size'} = $2;
		$flag1 = 1 ;
	    }
	    # recuperation du label du disque.
	    $hash->{$disk}{"label"} = qx($cmd_parted -s $disk p | grep -e "Partition Table:" -e "Table de partition :" 2>&1);
	    $hash->{$disk}{"label"} =~ s/.*: //;
	    chomp($hash->{$disk}{"label"});

	    $hash->{$disk}{"label"} = 'none' if (!$hash->{$disk}{"label"});
	    
	    # recuperation du model name
	    $hash->{$disk}{"model_name"} = _get_disk_model_name($disk);
	}
	# 255 headss, 63 sectors/track, 30392 cylinders
	elsif ($line =~ /^\s*(\d*)\s+.*,\s*(\d*)\s+sect.*,\s*(\d*).*/)
	{
	    $hash->{$disk}{'heads'} = $1;
	    $hash->{$disk}{'sectors'} = $2;
	    $hash->{$disk}{'cylinders'} = $3;
	    $flag2 = 1;
	}
	# Des que toutes les infos de fdisk sont recuperees, on recupere les infos de parted
	elsif ( $flag2 && $flag1)
	{
	    my @info_parted = qx!$cmd_parted -s $disk unit B print 2>&1!;
	    _get_parted_info($hash, \@info_parted, $disk, $version_parted);
	    $flag1 = 0;
	    $flag2 = 0;
	}
    }
}

# Test la presence et l'existence des commandes fdisk et parted
sub	_test_cmd
{
    my $cmd	= shift;
    my $name	= shift;
    my $func;
    
    chomp($func = `which $cmd`);
    
    if ( ! -e $func){
	$func = $name;
    }
    
    if ( ! -x $func){
	print "The command $cmd is not executable.\n";
	exit 0;
    }
    return $func;
}

# Cette methode test si le tableau passe en parametre contient
# une ligne commencant par Err*
# Il retourne le numero de la ligne + 1 contenant l'erreur
sub _test_parted_error
{
    my (@tab_parted) = @_;

    return (1) if (!@tab_parted);

    my $position = 1;
    foreach my $line (@tab_parted)
    {
	return ($position) if ($line =~ m/^(Error|Erreur): /);
	$position++;
    }
    return (0);
}

# Cette methode test si le tableau passe en parametre contient
# une ligne commencant par Warn* ou Aver*
sub _test_parted_warning
{
    my (@tab_parted) = @_;

    return (1) if (!@tab_parted);

    foreach my $line (@tab_parted)
    {
	return (1) if ($line =~ m/^(Avertissement|Warning): /);
    }
    return (0);
}

# Cette methode renvoie un bool et la regexp pour le traitement
# du retour de parted si la version de parted est connue
# Elle prend en parametre une string..
# DRY.....
sub _detect_parted_version
{
    my ($line) = @_;
    
    my $flag = 0;
    my $regexp = 0;
    
    if ($line =~ /\s+Start\s+End\s+Size\s+File\s+system\s+Name\s+Flags/ ||
	$line =~ /\s+D.but\s+Fin\s+Taille\s+Syst.me\sde\sfichiers\s+Nom\s+Fanions/)
    {
	$regexp	= $regex2;
	$flag	= 1;
    }
    elsif ($line =~ /\s+Start\s+End\s+Size\s+Type\s+File\s+system\s+Flags/)

    {
	$regexp	= $regex;
	$flag	= 1;
    }
    elsif ($line =~ /\s+Start\s+End\s+Size\s+File\s+system\s+Flags/ |
	   $line =~ /\s+D.but\s+Fin\s+Taille\s+Syst.me\sde\sfichiers\s+Fanions/)
    {
	$regexp	= $regex3;
	$flag	= 1;
    }
    # Si on a pas une des versions attendue, on affiche un message d'erreur et on quitte
    else
    {
	print $error_parted;
    }
    return $flag, $regexp;
}

# Fonction de recuperation du contenu d'un fichier...
sub _open_file_line
{
    my $path = shift;
    
    if (open (FILES, $path))
    {
	my $file = "";
	while (<FILES>)
	{
	    $file .= $_;
	}
	close FILES;
	return $file;
    }
    print ("ERROR $path: $!\n");
    
    return undef;
}

# Recuperation du model name d'un disque 
sub _get_disk_model_name
{
    my $dev = shift;

    return (0) if (!$dev);

    my $model_name = "N.A.";

    # Maintenant, on passe a la recuperation du nom du disque... la c sport
    if ($dev =~ m/\/hd.$/)
    {
	# recuperation du DD que l'on gere
	my $dd = ($dev =~ m/\/dev\/(.{3})/)[0];

 	# recuperation de la liste des DD
 	my $champ = qx(ls /proc/ide/ide*/*/model | grep $dd 2> /dev/null);

 	# recuperation du path pour le /proc correspondant a notre DD
 	$champ =~ s/model\n//;

 	# recuperation du model_name..
	chop (my $tmp = _open_file_line($champ . "model"));
	$model_name = $tmp if ($tmp);
    }
    ####
    # Gestion sata, usb, scsi..
    ####
    elsif ($dev =~ m/\/sd.$/)
    {
	# recuperation de la liste des infos pour le $device que l'on gere
	my @tmp_model = qx(scsi_info $dev);

	# boucling sur le tableau dernierement recupere
	foreach (@tmp_model)
	{
	    # Si je match "MODEL="
	    if ($_ =~ m/MODEL=\".*/)
	    {
		# je recupere le model_name
		$model_name = ($_ =~ m/MODEL=\"(.*)\"/)[0];
		
		# La ligne si dessous est incorrecte.. je laisse au cas ou
		# $model_name = [($_ =~ m/MODEL=\"(.*)\"/)[0], 0];
	    }
	}
    }
    return ($model_name);
}

1;
